@@ -1,13 +1,11 @@
 # -*- coding: utf-8 -*-
 """
-    # Shake.Auth
-
-    Copyright © 2010–2011 by Lúcuma labs (http://lucumalabs.com).
-    MIT License. (http://www.opensource.org/licenses/mit-license.php)
+Shake-Auth
 
 """
 from datetime import datetime
 import hashlib
+import hmac
 import time
 
 from jinja2.exceptions import TemplateNotFound
@@ -19,10 +17,19 @@ except ImportError:
     pass
 
 from .perms import REDIRECT_FIELD_NAME, is_valid
-from .utils import (bcrypt, LazyUser, Token,
-    auth_render, split_hash, hash_sha256, hash_sha512, hash_bcrypt)
+from .utils import *
+from .utils import bcrypt, scrypt, hash_old_sha256, hash_old_sha512
 
 
+__all__ = (
+    'Auth', 'CREDENTIAL_LOGIN', 'CREDENTIAL_PASSWORD', 'CREDENTIAL_TOKEN',
+    'UserExistsError', 'UserDoesNotExistsError', 'PasswordTooShortError'
+)
+
+CREDENTIAL_LOGIN = 'login'
+CREDENTIAL_PASSWORD = 'password'
+CREDENTIAL_TOKEN = 'token'
+
 class UserExistsError(ValueError):
     
     def __init__(self, login):
@@ -51,11 +58,6 @@ class PasswordTooShortError(ValueError):
             "%s chars long." % self.minlen)
 
 
-CREDENTIAL_LOGIN = 'login'
-CREDENTIAL_PASSWORD = 'password'
-CREDENTIAL_TOKEN = 'token'
-
-
 class Auth(object):
     """
     # Parameters:
@@ -133,12 +135,12 @@ class Auth(object):
     mailer
     :   a mailer instance (used in the reset password controller)
     
-    ..:note::
-    
+    ------------
+    NOTE:
     The passwords hashes are stored with the following
     format (without the spaces):
         
-        sha_name $ cost $ salt $ salted_and_hashed_password
+        dsha512 $ cost $ salt $ salted_and_hashed_password
         - or -
         bcrypt $ cost $ salt salted_and_hashed_password
 
@@ -147,7 +149,7 @@ class Auth(object):
     defaults = {
         'pepper': None,
         'hash_alg': None,
-        'hash_cost': 12,
+        'hash_cost': 11,
 
         'sign_in_redirect': '/',
         'sign_out_redirect': '/',
@@ -175,17 +177,26 @@ class Auth(object):
             settings.setdefault(k, default)
         
         assert settings.pepper is not None, "A `pepper` setting is required"
-        
+        settings.pepper = settings.pepper.encode('utf8', 'ignore')
+
         hash_alg = settings.hash_alg
         
         if not hash_alg:
-            if bcrypt:
+            if scrypt:
+                self.hash_func = hash_scrypt
+                settings.hash_alg = 'scrypt'
+            elif bcrypt:
                 self.hash_func = hash_bcrypt
                 settings.hash_alg = 'bcrypt'
             else:
                 self.hash_func = hash_sha512
                 settings.hash_alg = 'sha512'
         
+        elif hash_alg == 'scrypt':
+            if not scrypt:
+                raise ImportError("Hash algorithm 'scrypt' not available")
+            self.hash_func = hash_scrypt
+
         elif hash_alg == 'bcrypt':
             if not bcrypt:
                 raise ImportError("Hash algorithm 'bcrypt' not available")
@@ -198,12 +209,12 @@ class Auth(object):
             self.hash_func = hash_sha256
         
         else:
-            raise ValueError("The `hash_alg` parameter must be either " \
-                "'bcrypt', 'sha512' or 'sha256'")
+            raise AssertionError("The `hash_alg` parameter must be either " \
+                "'scrypt', 'bcrypt', 'sha512' or 'sha256'")
         
-        hash_cost = settings.hash_cost
-        assert isinstance(hash_cost, (int, long)) and hash_cost >= 0, \
-            "`hash_cost` parameter must be a positive integer"
+        hash_cost = settings.hash_cost or 4
+        assert isinstance(hash_cost, (int, long)) and hash_cost >= 4, \
+            "`hash_cost` parameter must be a positive integer bigger than 4"
         
         if (hash_alg == 'bcrypt') and (hash_cost < 4):
             settings.hash_cost = 4
@@ -228,6 +239,15 @@ class Auth(object):
             self.get_user(request)
     
 
+    def _get_user(self, login):
+        """Get user by login (case-insensitive).
+        """
+        login = login.lower()
+        lower_ = self.db.func.lower
+        return self.db.query(self.User) \
+            .filter(lower_(self.User.login) == login.lower()).first()
+
+
     def get_user(self, request):
         request.__class__.user = self.LazyUser(self.db, self.User)
     
@@ -236,24 +256,26 @@ class Auth(object):
         """Check if there's a matching pair of login/hashed_password in the
         database.
         
+        If `update` is True and the authentication was successful, update the
+        password hash to the current algorithm and cost.
+
         Returns the user if the credentials are correct, or None.
         """
         login = credentials.get(CREDENTIAL_LOGIN)
         password = credentials.get(CREDENTIAL_PASSWORD)
         if login and password:
+            login = login.encode('utf8', 'ignore')
+            password = password.encode('utf8', 'ignore')
             return self._loginpass_authentication(login, password, update)
         
         token = credentials.get(CREDENTIAL_TOKEN)
         if token:
+            token = token.encode('utf8', 'ignore')
             return self._token_authentication(token)
     
 
     def _loginpass_authentication(self, login, password, update):
-        # iOS by default makes you write the first letter as uppercase.
-        # The username should be case insensitive anyway.
-        login = login.lower()
-        user = self.db.query(self.User) \
-            .filter(self.User.login == login).first()
+        user = self._get_user(login)
         if not user:
             return None
         if not self.check_password(password, user.password):
@@ -263,11 +285,10 @@ class Auth(object):
             # If the authentication was successful, update the password hash
             # to the current algorithm and cost.
             hash_alg, hash_cost, salt = split_hash(user.password)
-            must_update_hash = hash_alg != self.settings.hash_alg
-            must_update_cost = hash_cost != self.settings.hash_cost
+            must_update_hash = (hash_alg != self.settings.hash_alg)
+            must_update_cost = (hash_cost != self.settings.hash_cost)
             if must_update_hash or must_update_cost:
-                new_password = self.hash_password(password)
-                user.password = new_password
+                user.password = self.hash_password(password)
         return user
     
 
@@ -277,8 +298,7 @@ class Auth(object):
         except ValueError:
             return None
         
-        user = self.db.query(self.User) \
-            .filter(self.User.login == login).first()
+        user = self._get_user(login)
         if not user:
             return None
         
@@ -296,8 +316,8 @@ class Auth(object):
     
 
     def hash_password(self, password):
-        return self.hash_func(password, cost=self.settings.hash_cost,
-            pepper=self.settings.pepper)
+        password += self.settings.pepper
+        return self.hash_func(password, cost=self.settings.hash_cost)
     
 
     def check_password(self, password, stored_password):
@@ -313,22 +333,32 @@ class Auth(object):
         except ValueError:
             raise ValueError("The stored password isn't a valid hash")
         
+        password += self.settings.pepper
+
+        if hash_alg == 'scrypt':
+            assert 'scrypt' in globals().keys(), ('\'scrypt\' is not '
+                'available but the stored password was hashed using it.\n '
+                'Please install scrypt '
+                '<http://pypi.python.org/pypi/scrypt/>')
+            hashed = hash_scrypt(password, stored_password, hash_cost)
+
         if hash_alg == 'bcrypt':
             assert 'bcrypt' in globals().keys(), ('\'bcrypt\' is not '
                 'available but the stored password was hashed using it.\n '
                 'Please install py-bcrypt '
                 '<http://pypi.python.org/pypi/py-bcrypt/>')
-            
-            hashed = hash_bcrypt(password, hash_cost, stored_password,
-                self.settings.pepper)
+
+        elif hash_alg == 'dsha512':
+            hashed = hash_sha512(password, salt, hash_cost)
         
+        elif hash_alg == 'dsha256':
+            hashed = hash_sha256(password, salt, hash_cost)
+
         elif hash_alg == 'sha512':
-            hashed = hash_sha512(password, hash_cost, salt,
-                self.settings.pepper)
-        
+            hashed = hash_old_sha512(password, salt, hash_cost)
+
         elif hash_alg == 'sha256':
-            hashed = hash_sha256(password, hash_cost, salt,
-                self.settings.pepper)
+            hashed = hash_old_sha256(password, salt, hash_cost)
         
         else:
             raise ValueError("Hash algorithm '%s' not available" % hash_alg)
@@ -336,6 +366,7 @@ class Auth(object):
     
 
     def set_password(self, user, passw):
+        passw = passw.encode('utf8', 'ignore')
         if len(passw) < self.settings.password_minlen:
             raise PasswordTooShortError(self.settings.password_minlen)
         user.password = self.hash_password(passw)
@@ -354,16 +385,16 @@ class Auth(object):
         fake tokens cannot be generated even if the database is compromised.
         """
         time36 = time36 or to36(int(time.time()))
-        msg = ''.join([
-            self.app.settings.SECRET_KEY,
+        key = ''.join([
+            self.app.settings.secret_key,
             getattr(user, 'password', '')[10:50],
             unicode(getattr(user, 'last_sign_in', '')),
             unicode(getattr(user, 'id', 0)),
             unicode(time36),
         ])
-        
-        mac = hashlib.sha512(msg).hexdigest()[:50]
-        
+        key = key.encode('utf8', 'ignore')
+        mac = hmac.new(key, msg=None, digestmod=hashlib.sha512)
+        mac = mac.hexdigest()[:50]
         token = '%s.%s.%s' % (user.login, time36, mac)
         return Token(token, self.settings.reset_expire)
     
@@ -375,15 +406,15 @@ class Auth(object):
     def sign_out(self, request):
         self.LazyUser.sign_out(request)
     
-    # This method is intended to be called from the command line.
+
     def create_user(self, login, passw, **data):
         """[-login] USER_NAME [-passw] PASSWORD
-        Creates a new user."""
-        # iOS by default makes you write the first letter as uppercase.
-        # The username should be case insensitive anyway.
-        login = login.lower()
-        user_exists = self.db.query(self.User) \
-            .filter(self.User.login == login).count()
+        Creates a new user.
+        This method is intended to be called from the command line.
+        """
+        login = login.encode('utf8', 'ignore')
+        passw = passw.encode('utf8', 'ignore')
+        user_exists = self._get_user(login)
         if user_exists:
             raise UserExistsError(login)
         if len(passw) < self.settings.password_minlen:
@@ -395,15 +426,15 @@ class Auth(object):
         self.db.commit()
         return user
     
-    # This method is intended to be called from the command line.
+    
     def change_password(self, login, passw):
         """[-login] USER_NAME [-passw] NEW_PASSWORD
-        Changes the password of an existing user."""
-        # iOS by default makes you write the first letter as uppercase.
-        # The username should be case insensitive anyway.
-        login = login.lower()
-        user = self.db.query(self.User) \
-            .filter(self.User.login == login).first()
+        Changes the password of an existing user.
+        This method is intended to be called from the command line.
+        """
+        login = login.encode('utf8', 'ignore')
+        passw = passw.encode('utf8', 'ignore')
+        user = self._get_user(login)
         if not user:
             raise UserDoesNotExistsError(login)
         self.set_password(user, passw)
@@ -508,12 +539,8 @@ class Auth(object):
         
         # Phase 2
         elif request.is_post:
-            login = request.form.get('login')
-            # iOS by default makes you write the first letter as uppercase.
-            # The username should be case insensitive anyway.
-            login = login.lower()
-            user = self.db.query(self.User) \
-                .filter(self.User.login == login).first()
+            login = request.form.get('login', '')
+            user = self._get_user(login)
             if not user:
                 kwargs['error'] = 'NOT FOUND'
             else:
